{% if is_first_by_category %}
##
## enum types
##

{% endif%}
class {{ metadata.classname }}(object):
    """
    {{ metadata.docs }}
    """
    {% for value_name in metadata.values %}
    {{ value_name }}: int = {{ metadata.values[value_name].value }}
    """
    {{ metadata.values[value_name].docs }}
    """
    {% endfor %}
