# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

class Serializer(object):
    TRANSPORT = 0
    JSON = 1
    MSGPACK = 2
    CBOR = 3
    UBJSON = 4
    OPAQUE = 5
    FLATBUFFERS = 6

