# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

class AuthMethod(object):
    ANONYMOUS = 0
    COOKIE = 1
    TLS = 2
    TICKET = 3
    CRA = 4
    SCRAM = 5
    CRYPTOSIGN = 6

